from __future__ import annotations

from datetime import date, datetime
from typing import Annotated, Literal

from pydantic import BeforeValidator
from pydantic.dataclasses import Field, dataclass

from .common import as_float
from .session import session


def convert_to_datetime(x: str | datetime) -> datetime:
    if isinstance(x, datetime):
        if x.tzinfo is None:
            raise ValueError("No timezone information found in this date")
        return x

    try:
        return datetime.strptime(x, "%Y-%m-%dT%H:%M:%S%z")
    except ValueError:  # 2023-11-16 08:24
        return datetime.strptime(x, "%Y-%m-%d %H:%M")


def convert_to_date(x: str | date | datetime) -> date:
    if isinstance(x, datetime):
        return x.date()
    if isinstance(x, date):
        return x

    return datetime.strptime(x, "%Y-%m-%d").date()


Url = Annotated[str, BeforeValidator(lambda x: session.create_url(x))]
Date = Annotated[date, BeforeValidator(convert_to_date)]
DateTime = Annotated[datetime, BeforeValidator(convert_to_datetime)]


@dataclass
class CourseGraphic:
    type: Literal["icon"]
    value: str


@dataclass
class ResultGraphic:
    type: Literal["percentage"]
    color: Literal["green", "red"]
    value: int
    description: str

    @property
    def achieved_points(self) -> float:
        return as_float(self.description.split("/")[0])

    @property
    def total_points(self) -> float:
        return as_float(self.description.split("/")[1])


@dataclass
class PersonDescription:
    startingWithFirstName: str = ""
    startingWithLastName: str = ""


@dataclass
class _User:
    id: str
    pictureHash: str
    pictureUrl: Url
    description: PersonDescription
    name: PersonDescription
    sort: str


@dataclass
class DateRange:
    start: DateTime
    end: DateTime


@dataclass
class SkoreWorkYear:
    id: int
    dateRange: DateRange


@dataclass
class Class_:
    identifier: str
    id: int
    platformId: int
    name: str
    type: str
    icon: str


@dataclass
class Period:
    id: int
    name: str
    icon: str
    skoreWorkYear: SkoreWorkYear
    isActive: bool
    class_: Class_ = Field(alias="class")


@dataclass
class Component:
    id: int
    name: str
    abbreviation: str


Teacher = _User
Student = _User


@dataclass
class Course:
    id: int
    name: str
    graphic: CourseGraphic
    teachers: list[Teacher]
    skoreClassId: int
    parentCourseId: None | int
    skoreWorkYear: SkoreWorkYear
    class_: Class_ = Field(alias="class")


@dataclass
class Feedback:
    text: str
    user: Teacher


@dataclass
class FeedbackFull:
    attachments: list[str]
    changedAt: DateTime
    createdAt: DateTime
    evaluationId: str
    id: str
    student: Student
    teacher: Teacher
    text: str


@dataclass
class Result:
    identifier: str
    type: Literal["normal"]
    name: str
    graphic: ResultGraphic
    date: DateTime

    gradebookOwner: Teacher
    component: Component
    courses: list[Course]
    period: Period
    feedback: list[Feedback]
    feedbacks: list[FeedbackFull]
    availabilityDate: DateTime
    isPublished: bool
    doesCount: bool


@dataclass
class ResultDetails:
    centralTendencies: list[str]
    teachers: list[Teacher]
    dateChanged: DateTime
    userChanged: Teacher
    class_: Class_ = Field(alias="class")


@dataclass
class ResultWithDetails(Result):
    details: ResultDetails


@dataclass
class CourseCondensed:
    name: str
    teacher: str
    url: Url

    descr: str = Field(repr=False, default="")
    icon: str = Field(repr=False, default="")


@dataclass
class FutureTaskOneTask:
    label: str
    description: str
    icon: str
    warning: bool
    click_handle: str
    activityID: int
    dateID: str
    assignmentID: str
    endMomentTS: None
    startMomentID: str
    endMomentID: str
    lessonID: str
    type: str
    classID: str
    course: str
    date: Date
    hourID: str


@dataclass
class FutureTaskOneItem:
    tasks: list[FutureTaskOneTask]
    materials: list[str]


@dataclass
class FutureTaskOneCourse:
    lessonID: str
    hourID: str
    classID: str
    course_title: str
    items: FutureTaskOneItem


@dataclass
class FutureTaskOneDay:
    date: Date
    pretty_date: str
    courses: list[FutureTaskOneCourse]


@dataclass
class FutureTasks:
    """
    Class that interfaces the retrieval of any task that needs to be made in the near future.

    Example:
    -------
    >>> for day in FutureTasks().days:
    >>>     for course in day.courses:
    >>>         print("Course:", course.course_title)
    >>>         for item in course.items:
    >>>             for task in item.tasks:
    >>>                 print("Task:", task.description)
    Course: 2 - AAR1, Lotte Peeters
    Task: Toets 3. De koolstofcyclus in het systeem aarde pagina 42 - 47
    """

    days: list[FutureTaskOneDay] = Field(default_factory=list)
    last_assignment_id: int = 0
    last_date: Date = Field(default_factory=date.today)

    def __post_init__(self):
        """I need to do this here because when I do it in Agenda, it'll not lazily load it. But in this way, I load it on construction."""
        json = session.json("/Agenda/Futuretasks/getFuturetasks")

        self.days = []
        for d in json["days"]:
            self.days.append(FutureTaskOneDay(**d))

        self.last_assignment_id = json["last_assignment_id"]
        self.last_date = convert_to_date(json["last_date"])


@dataclass
class AgendaHour:
    hourID: str
    start: str
    end: str
    title: str


@dataclass
class AgendaLesson:
    momentID: str
    lessonID: str
    hourID: str
    date: Date
    subject: str | None
    course: str
    courseTitle: str
    classroom: str
    classroomTitle: str
    teacher: str
    teacherTitle: str
    klassen: str
    klassenTitle: str
    classIDs: str
    bothStartStatus: str
    assignmentEndStatus: str
    testDeadlineStatus: str
    noteStatus: str
    note: str | None
    date_listview: str
    hour: str
    activity: str
    activityID: str | None
    color: str
    hourValue: str
    components_hidden: object
    freedayIcon: str
    someSubjectsEmpty: str | None

    @property
    def hour_details(self) -> AgendaHour:
        from .agenda import SmartschoolHours

        return SmartschoolHours().search_by_hourId(self.hourID)


@dataclass
class AgendaMomentInfoAssignment:
    startAssignment: str
    start: str
    end: str
    type: str
    description: str
    atdescription: str
    freedeadline: str
    warning: str
    assignmentInfo: str
    assignmentDeadline: str


@dataclass
class AgendaMomentInfo:
    className: str
    subject: str
    materials: str | None
    momentID: str
    assignments: list[AgendaMomentInfoAssignment]


@dataclass
class StudentSupportLink:
    id: str
    name: str
    description: str
    icon: str
    link: Url
    cleanLink: str
    isVisible: bool


@dataclass
class Message:
    id: int
    fromImage: Url
    subject: str
    date: DateTime
    status: int
    attachment: bool
    unread: bool
    label: bool
    deleted: bool
    allowreply: bool
    allowreplyenabled: bool
    hasreply: bool
    hasForward: bool
    realBox: str
    sendDate: DateTime | None

    from_: str = Field(alias="from")



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import date
from typing import Iterator, TypeVar
from xml.etree import ElementTree as ET
from xml.sax.saxutils import quoteattr

from .common import xml_to_dict
from .session import session

_T = TypeVar("_T")


class SmartschoolXML(ABC):
    __is_cached__ = True

    def _construct_command(self) -> str:
        txt = "<request><command>"
        txt += f"<subsystem>{self._subsystem}</subsystem>"
        txt += f"<action>{self._action}</action>"
        txt += "<params>"

        for k, v in self._params.items():
            txt += f'<param name="{quoteattr(k)}"><![CDATA[{v}]]></param>'

        txt += "</params></command></request>"
        return txt

    def __iter__(self) -> Iterator[_T]:
        yield from self._xml()

    def _xml(self, date_to_use: date | None = None):
        today = date_to_use or date.today()
        current_week = today.strftime("%Y-%U")
        if self.__is_cached__ and current_week in self.cache:
            return self.cache[current_week]

        response = session.post(
            "/?module=Agenda&file=dispatcher",
            data={"command": self._construct_command()},
            headers={
                "X-Requested-With": "XMLHttpRequest",
            },
        )

        root = ET.fromstring(response.text)

        all_entries = []
        as_obj = self._object_to_instantiate
        for el in root.findall(self._xpath):
            as_dict = xml_to_dict(el)
            self._post_process_element(as_dict)
            obj = as_obj(**as_dict)
            all_entries.append(obj)

        self.cache[current_week] = all_entries

        return all_entries

    @property
    def cache(self) -> dict:
        if self.__is_cached__:
            raise NotImplementedError("You should add a `cache: ClassVar[dict]` to your derived class.")
        return {}

    @property
    @abstractmethod
    def _subsystem(self) -> str:
        """Returns the subsystem to request the info from."""

    @property
    @abstractmethod
    def _action(self) -> str:
        """Returns the action to send."""

    @property
    @abstractmethod
    def _params(self) -> dict:
        """Returns the parameters to send."""

    @property
    @abstractmethod
    def _xpath(self) -> str:
        """Returns the xpath to investigate."""

    @property
    @abstractmethod
    def _object_to_instantiate(self) -> type[_T]:
        """Returns the object to instantiate."""

    def _post_process_element(self, element: dict) -> None:  # noqa: B027
        """By default, this doesn't do anything, but you can adjust it when needed."""



from __future__ import annotations

import contextlib
import functools
import inspect
import json
import operator
import platform
import re
import smtplib
import sys
import traceback
import xml.etree.ElementTree as ET
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from enum import Enum, auto
from pathlib import Path
from typing import Callable, Literal

from bs4 import BeautifulSoup, FeatureNotFound
from requests import Response

CACHE = Path(__file__).parent / ".cache"


__all__ = [
    "send_email",
    "CACHE",
    "capture_and_email_all_exceptions",
    "save",
    "IsSaved",
    "bs4_html",
    "get_all_values_from_form",
    "make_filesystem_safe",
    "as_float",
    "xml_to_dict",
]

_used_bs4_option = None


class IsSaved(Enum):
    NEW = auto()
    UPDATED = auto()
    SAME = auto()


def save(
    type_: Literal["agenda", "punten", "todo"], course_name: str, id_: str, data: dict | str, is_eq: Callable = operator.eq, extension: str = "json"
) -> IsSaved | dict:
    save_as = CACHE / f"_{type_}/{course_name}/{id_}.{extension}"
    save_as.parent.mkdir(exist_ok=True, parents=True)
    data_was_dict = isinstance(data, dict)

    if data_was_dict:
        to_write = json.dumps(data, indent=4)
    else:
        to_write = data

    if not save_as.exists():
        save_as.write_text(to_write, encoding="utf8")
        return IsSaved.NEW

    old_data = save_as.read_text(encoding="utf8")
    if data_was_dict:
        old_data = json.loads(old_data)

    if is_eq(data, old_data):
        return IsSaved.SAME

    save_as.write_text(to_write, encoding="utf8")
    return old_data


def send_email(
    subject: str,
    text: str,
    email_to: list[str] | str,
    email_from: str,
):
    if isinstance(email_to, str):
        email_to = [email_to]

    message = MIMEMultipart("alternative")
    message["Subject"] = subject
    message["From"] = email_from
    message["To"] = ", ".join(email_to)
    message.attach(MIMEText(text, "plain", "utf8"))

    print(f"Sending email >> {subject}")

    if platform.system() == "Windows":
        print("=================== On Linux we would have sent this: ===================")
        print(message.as_string())
        print("=========================================================================")
        return  # I put this here, so I can still debug 'message.as_string()'

    with smtplib.SMTP("localhost") as server:
        server.sendmail(
            from_addr=email_from,
            to_addrs=email_to,
            msg=message.as_string(),
        )


def capture_and_email_all_exceptions(func, email_from: str | list[str], email_to: str | list[str]) -> Callable:
    @functools.wraps(func)
    def inner(*args, **kwargs):
        frm = inspect.stack()[1]
        module_name = Path(frm.filename)
        function_signature = f"{module_name.stem}.{func.__name__}"

        print(f"[{function_signature}] Start")
        try:
            result = func(*args, **kwargs)
        except Exception as ex:
            print(f"[{function_signature}] An exception happened: {ex}")

            send_email(
                email_to=email_to,
                email_from=email_from,
                subject="Salco parser: something went wrong!!",
                text="".join(traceback.format_exception(None, ex, ex.__traceback__)),
            )

            sys.exit(1)

        print(f"[{function_signature}] Finished")
        return result

    return inner


def bs4_html(html: str | bytes | Response) -> BeautifulSoup:
    global _used_bs4_option

    if isinstance(html, Response):
        html = html.text

    possible_options = [
        _used_bs4_option,
        {"parser": "html.parser", "features": "lxml"},
        {"features": "html5lib"},
        {"features": "html.parser"},
    ]

    for kw in possible_options:
        if kw is None:
            continue

        with contextlib.suppress(FeatureNotFound):
            parsed = BeautifulSoup(html, **kw)
            _used_bs4_option = kw
            return parsed

    _used_bs4_option = {}
    return BeautifulSoup(html)


def get_all_values_from_form(html, form_selector):
    form = html.select(form_selector)
    assert len(form) == 1, f"We should have only 1 form. We got {len(form)}!"
    form = form[0]

    # action = form.attrs.get("action").lower()
    # method = form.attrs.get("method", "get").lower()
    all_inputs = form.find_all(["input", "button", "textarea", "select"])

    inputs = []
    for input_tag in all_inputs:
        tag_name = input_tag.name.lower()
        attrs = input_tag.attrs

        if "name" not in attrs:
            continue

        if tag_name != "select":
            inputs.append(
                {
                    "name": attrs.get("name"),
                    "value": attrs.get("value", ""),
                }
            )
        else:  # select
            raise ValueError("Check if this code works. Possible issue with getting the value if the value tag isn't set.")

            select_options = []
            value = ""
            for select_option in input_tag.find_all("option"):
                option_value = select_option.attrs.get("value")
                if option_value:
                    select_options.append(option_value)
                    if "selected" in select_option.attrs:
                        value = option_value
            if not value and select_options:
                # if the default is not set, and there are options, take the first option as default
                value = select_options[0]
            # add the select to the inputs list
            inputs.append({"name": attrs.get("name"), "values": select_options, "value": value})

    return inputs


def make_filesystem_safe(name: str) -> str:
    return re.sub("[^-_a-z0-9.]+", "", name, flags=re.IGNORECASE)


def as_float(txt: str) -> float:
    return float(txt.replace(",", "."))


def xml_to_dict(element, *, depth: int = 0):
    if depth == 0 and isinstance(element, str):
        element = ET.XML(element.strip())

    result = {}

    for child in element:
        tag = child.tag
        if len(child) == 0:  # Leaf node?
            child_data = child.text
        else:
            child_data = xml_to_dict(child)

        if tag in result:
            if not isinstance(result[tag], list):
                # If the tag already exists, convert it to a list
                result[tag] = [result[tag]]
            result[tag].append(child_data)
        else:
            result[tag] = child_data

    return result



from __future__ import annotations

from enum import Enum

from ._xml_interface import SmartschoolXML
from .objects import Message

__all__ = ["SortField", "SortOrder", "BoxType", "Messages"]


class SortField(Enum):
    DATE = "date"
    FROM = "from"
    READ_UNREAD = "status"
    ATTACHMENT = "attachment"
    FLAG = "label"


class SortOrder(Enum):
    ASC = "asc"
    DESC = "desc"


class BoxType(Enum):
    INBOX = "inbox"
    DRAFT = "draft"
    SCHEDULED = "scheduled"
    SENT = "outbox"
    TRASH = "trash"


class Messages(SmartschoolXML):
    """
    Interfaces the mailbox principle in Smartschool.

    To reproduce: Open the messages and watch the XHR calls.

    Example:
    -------
    >>> for message in Messages():
    >>>     print(message.subject)
    Re: LO les
    Frans
    """

    __is_cached__ = False

    def __init__(
        self,
        box_type: BoxType = BoxType.INBOX,
        sort_by: SortField = SortField.DATE,
        sort_order: SortOrder = SortOrder.DESC,
        already_seen_message_ids: list[int] | None = None,
    ):
        self._box_type = box_type
        self._sort_by = sort_by
        self._sort_order = sort_order
        self._already_seen_message_ids = already_seen_message_ids or []

    @property
    def _subsystem(self) -> str:
        return "postboxes"

    @property
    def _action(self) -> str:
        return "message list"

    @property
    def _params(self) -> dict:
        return {
            "boxType": self._box_type.value,
            "boxID": "0",
            "sortField": self._sort_by.value,
            "sortKey": self._sort_order.value,
            "poll": "false" if not self._already_seen_message_ids else "true",
            "poll_ids": ",".join(str(x) for x in self._already_seen_message_ids),
            "layout": "new",
        }

    @property
    def _xpath(self) -> str:
        return ".//messages/message"

    @property
    def _object_to_instantiate(self) -> type[Message]:
        return Message
